shader_type particles;

uniform bool      centered             = true;
uniform vec2      size                 = vec2(1024.);
uniform float     texture_scale        = 1.0;
uniform float     texture_rot_speed    = .0;
uniform float     texture_scale_rnd : hint_range(.0, 1.) = .0;
uniform float     wind                 = .0;
uniform float     wind_influence       = 24.;
uniform float     gravity              = 5.;
uniform float     time_scale           = 1.0;
uniform float     randomization        = 1.0;
uniform float     rotspeed1            = 4.0;
uniform float     radius1              = 64.0;
uniform float     rotspeed2            = 10.;
uniform float     radius2              = 50.0;
uniform vec3      alpha_ramp           = vec3(.02, .5, .8);
uniform float     alpha_value          = 1.0;
uniform int       trail_divisor        = 1;

uniform bool test0;
uniform bool test1;
uniform bool test2;
uniform bool test3;
uniform bool test4;
uniform bool test5;
uniform bool test6;

const float PI = 3.14159;
const float TWO_PI = PI * 2.0;
const float SEMI_PI = PI / 2.0;

float rand_from_seed(inout uint seed) {
	int k;
	int s = int(seed);
	if (s == 0)
	s = 305420679;
	k = s / 127773;
	s = 16807 * (s - k * 127773) - 2836 * k;
	if (s < 0)
		s += 2147483647;
	seed = uint(s);
	return float(seed % uint(65536)) / 65535.0;
}

float rand_from_seed_m1_p1(inout uint seed) {
	return rand_from_seed(seed) * 2.0 - 1.0;
}

uint hash(uint x) {
	x = ((x >> uint(16)) ^ x) * uint(73244475);
	x = ((x >> uint(16)) ^ x) * uint(73244475);
	x = (x >> uint(16)) ^ x;
	return x;
}

void rotate(inout mat4 transform, float angle) {
	transform[0] = vec4(cos(angle), -sin(angle), 0.0, 0.0);
	transform[1] = vec4(sin(angle), cos(angle), 0.0, 0.0);
	transform[2] = vec4(0.0, 0.0, 1.0, 0.0);
}

void resize(inout mat4 transform, float q) {
	vec3 real_scale = vec3(texture_scale * q);
	transform[0].xyz *= real_scale;
	transform[1].xyz *= real_scale;
	transform[2].xyz *= real_scale;
}

void update_alpha(inout vec4 color, float value) {
	color.a = alpha_value * (
			smoothstep(alpha_ramp.x, alpha_ramp.y, value)
			- smoothstep(alpha_ramp.z, 1.0, value)
			);
}

vec2 get_random_position(uint random_seed) {
	vec2 r = vec2(
			rand_from_seed(random_seed) * size.x,
			rand_from_seed(random_seed) * size.y);
	if (centered) { r -= size / 2.f; }
	return r;
}

void init_particle(uint base_number, uint random_seed, inout vec4 custom, mat4 emission_transform, inout mat4 transform, inout vec3 velocity) {
	vec2 pos = get_random_position(random_seed) + emission_transform[3].xy;
	custom.x = rand_from_seed(random_seed) * randomization + 1.;
	if (rand_from_seed(random_seed) > .4) { custom.x *= -1.; }
	custom.y = .0;
	custom.z = pos.x; // - wind * wind_influence * 10.;
	custom.w = pos.y;
	// Aleatoricidad de escala. TRANSFORM[3] no es para almacenar info.. pero ya que no esta usandose por ser 2d, lo utilizo xDDD
	transform[3].z = 1.0;//(texture_scale_rnd * rand_from_seed(random_seed));
}

void vertex() {
	uint base_number = uint(0);
	if (test5 == false) base_number = NUMBER / uint(trail_divisor);
	uint random_seed = uint(0);
	if (test6 == false) random_seed = hash(base_number + uint(1) + RANDOM_SEED);
	if (RESTART) {
		if (test0 == false) {
			init_particle(
					base_number,
					random_seed,
					CUSTOM,
					EMISSION_TRANSFORM,
					TRANSFORM,
					VELOCITY
					);
		}
	} else {
		CUSTOM.y += DELTA / LIFETIME;
	}
	float t = CUSTOM.y * time_scale;
	if (test1 == false) {
		float angle = atan(TRANSFORM[1].x, TRANSFORM[1].y);
		rotate(TRANSFORM, angle + texture_rot_speed * CUSTOM.x * DELTA);
		resize(TRANSFORM, texture_scale * (1. + TRANSFORM[3].z));
		update_alpha(COLOR, CUSTOM.y);
		float rnd = fract(CUSTOM.y) * .1; //rand_from_seed(random_seed);
	}
	if (test2 == false) {
		CUSTOM.z += wind * wind_influence * DELTA;
		CUSTOM.w += gravity * DELTA;
	}
	float rotspeed = rotspeed1 * CUSTOM.x;
	if (test3 == false) {
		float radius   = radius1 * CUSTOM.x;
		TRANSFORM[3].x = CUSTOM.z + cos(t * rotspeed) * radius;
		TRANSFORM[3].y = CUSTOM.w + sin(t * rotspeed) * radius;
		rotspeed = rotspeed2 * CUSTOM.x;
	}
	if (test4 == false) {
		TRANSFORM[3].x += cos((t) * rotspeed) * radius2;
		TRANSFORM[3].y += sin((t) * rotspeed) * radius2;
	}
}